from nltk import tokenize
from .string import strip_non_alphanumeric_from_beginning
from .string import ends_with_punctuation

class String:
	def __init__(self, string, font_size):
		self._string = strip_non_alphanumeric_from_beginning(string)
		self._string = ' '.join(self._string.split())
		self._font_size = font_size

	@property
	def font_size(self):
		return self._font_size

	@property
	def string(self):
		"""
		:rtype: str
		"""
		return self._string

	def __str__(self):
		return self._string

	def __repr__(self):
		return str(self)


class Title(String):
	def __str__(self):
		return f'* {self._string} *'


class Paragraph:
	def __init__(self, language='english'):
		self._language = language
		self._sentences = []
		self._num_words = 0

	@property
	def num_words(self):
		"""
		:rtype: int
		"""
		return self._num_words

	def append(self, sentence, num_words=None):
		num_words = num_words or len(tokenize.word_tokenize(sentence, language=self._language))
		self._sentences.append(sentence)
		self._num_words += num_words

	@property
	def string(self):
		"""
		:rtype: str
		"""
		return ' '.join(self._sentences)

	def __str__(self):
		return self.string

	def __repr__(self):
		return str(self)


class Text(String):
	def __init__(self, string, font_size, paragraph_max=50, language='english'):
		super().__init__(string=string, font_size=font_size)
		self._paragraph_max = paragraph_max
		self._language = language
		self._paragraphs = None

	@property
	def sentences(self):
		"""
		:rtype: list[str]
		"""
		return [
			strip_non_alphanumeric_from_beginning(sentence)
			for sentence in tokenize.sent_tokenize(self._string, language=self._language)
		]

	@property
	def paragraphs(self):
		"""
		:rtype: list[str]
		"""
		sentences = self.sentences
		num_sentences = len(sentences)

		if self._paragraphs is None:
			paragraphs = []
			for sentence in sentences:
				num_words = len(tokenize.word_tokenize(sentence, language=self._language))
				if not ends_with_punctuation(sentence) and num_sentences < 2 and (num_words < 8 or len(sentence) < 12):
					break
				if len(paragraphs) == 0:
					new_paragraph = Paragraph()
					new_paragraph.append(sentence=sentence, num_words=num_words)
					paragraphs.append(new_paragraph)
				else:
					last_paragraph = paragraphs[-1]
					if num_words + last_paragraph.num_words > self._paragraph_max - 2:
						new_paragraph = Paragraph()
						new_paragraph.append(sentence=sentence, num_words=num_words)
						paragraphs.append(new_paragraph)
					else:
						last_paragraph.append(sentence=sentence, num_words=num_words)
			self._paragraphs = [paragraph.string for paragraph in paragraphs]
		return self._paragraphs



from disk import Path
from joblib import Parallel
from joblib import delayed
from pandas import concat
from pandas import DataFrame
from chronometry.progress import iterate

from .PdfFile import PdfFile


class PdfDirectory(Path):
	def __init__(self, string, html_directory):
		super().__init__(string=string, show_size=False)
		self._html_directory = Path(html_directory)
		self._html_directory.make_dir(ignore_if_exists=True)
		self._pdfs = [
			PdfFile(x, html_directory=self._html_directory)
			for x in self.list(show_size=False) if x.extension.lower() == 'pdf'
		]
		self._errors = []

		self._data = None

	def reset_errors(self):
		self._errors = []

	@property
	def pdfs(self):
		"""
		:rtype: PdfFile
		"""
		return self._pdfs

	def get_data(self, num_jobs=1, **kwargs):
		def _get_data(pdf):
			"""
			:type pdf: PdfFile
			:rtype: DataFrame
			"""
			try:
				return pdf.get_data(**kwargs)
			except Exception as e:
				self._errors.append({'exception': e, 'pdf': pdf})
				return None

		if num_jobs == 1:
			dataframes = [
				_get_data(pdf=pdf)
				for pdf in iterate(self.pdfs, text='getting the data')
			]
		else:
			processor = Parallel(n_jobs=num_jobs, backend='threading', require='sharedmem')
			dataframes = processor(
				delayed(_get_data)(pdf=pdf)
				for pdf in iterate(self.pdfs, text='getting the data in parallel')
			)
		self._data = concat([x for x in dataframes if x is not None])
		if len(self._errors) > 0:
			raise self._errors[0]['exception']

	def get_time_data(self, num_jobs=1):
		"""
		:rtype: DataFrame
		"""
		def _get_time_data(x):
			return dict(pdf=x.name, **x._elapsed)

		if num_jobs == 1:
			records = [
				_get_time_data(pdf)
				for pdf in iterate(self.pdfs, text='getting processing time')
			]
		else:
			processor = Parallel(n_jobs=num_jobs, backend='threading', require='sharedmem')
			records = processor(
				delayed(_get_time_data)(pdf)
				for pdf in iterate(self.pdfs, text='getting processing time in parallel')
			)

		return DataFrame.from_records(records)



from disk import Path
from silverware import HtmlFile, Spoon
from pandas import DataFrame
from chronometry import get_elapsed
from chronometry import get_now

from .convert_pdf_to_html import convert_pdf_to_html
from .HtmlText import HtmlText
from .Text import Title
from .Text import Text


class PdfFile(Path):
	def __init__(self, string, vocabulary=None, show_size=False, html_directory=None):
		super().__init__(string=string, show_size=show_size)
		self._html_file = None
		self._spoon = None
		self._paragraphs = None
		self._vocabulary = vocabulary

		if html_directory is not None:
			html_directory = Path(html_directory)
		self._html_directory = html_directory

		self._elapsed = {}

	@property
	def html_file(self):
		"""
		:rtype: HtmlFile
		"""
		if self._html_file is None:
			self._html_file = self.convert_to_html()
		return self._html_file

	def get_html_path(self, path=None):
		"""
		:type path: NoneType or str or Path or HtmlPath
		:rtype: HtmlFile
		"""
		if path is None:
			if self._html_directory is None:
				html_directory = self.parent_directory + 'html'
			else:
				html_directory = self._html_directory
			html_directory.make_dir(ignore_if_exists=True)
			path = HtmlFile(string=(html_directory + (self.name + '.html')).path)
		elif isinstance(path, HtmlFile):
			pass
		elif isinstance(path, Path):
			path = HtmlFile(path)
		elif isinstance(path, str):
			if '.' in path:
				path = HtmlFile(path)
			else:
				path = HtmlFile(path + '.html')
		else:
			raise TypeError(f'html_path is of type {type(path)}!')
		return path

	def convert_to_html(self, path=None, ignore_if_exists=True, num_tries=3):
		"""
		:rtype: HtmlFile
		"""
		path = self.get_html_path(path=path)

		start_time = get_now()
		for try_num in range(num_tries + 1):
			if path.exists() and ignore_if_exists and path.get_size_kb() > 6:
				break
			else:
				convert_pdf_to_html(pdf_path=self, html_path=path)
				if path.exists() and path.get_size_kb() > 6:
					self._elapsed['html_conversion'] = get_elapsed(start=start_time, unit='s')
					break
		else:
			raise RuntimeError(f'could not convert {self.name_and_extension} to html after {num_tries} tries')

		self._html_file = path
		return path

	@property
	def spoon(self):
		"""
		:rtype: Spoon
		"""
		if self._spoon is None:
			start_time = get_now()
			self._spoon = self.html_file.spoon
			self._elapsed['html_parsing'] = get_elapsed(start=start_time, unit='s')
		return self._spoon

	def get_text(
			self, simplify=True, ignore_depth=True, flatten=True, remove_short_strings=False, remove_tables=True,
			min_english_ratio=0
	):
		"""
		:rtype: HtmlText
		"""
		soup = self.spoon.soup

		start_time = get_now()
		html_text = HtmlText(tag=soup, vocabulary=self._vocabulary)
		self._elapsed['text_generation'] = get_elapsed(start=start_time, unit='s')
		start_time = get_now()
		if simplify:
			html_text = html_text.simplify(
				ignore_depth=ignore_depth,
				remove_short_strings=False,
				remove_tables=remove_tables,
				min_english_ratio=min_english_ratio
			)
		if flatten:
			html_text = html_text.flatten()
		if simplify:
			html_text = html_text.simplify(
				ignore_depth=ignore_depth,
				remove_short_strings=remove_short_strings,
				remove_tables=remove_tables,
				min_english_ratio=min_english_ratio
			)
		self._elapsed['simplification'] = get_elapsed(start=start_time, unit='s')
		return html_text

	def get_text_records(
			self, simplify=True, ignore_depth=True, remove_short_strings=False,
			paragraph_max=50, min_english_ratio=0
	):
		"""
		:rtype: list[dict[str, ]]
		"""

		try:
			text = self.get_text(
				simplify=simplify, ignore_depth=ignore_depth,
				flatten=True, remove_short_strings=remove_short_strings,
				min_english_ratio=min_english_ratio
			)
		except Exception as e:
			print(f'exception raised when tried to get_text for {self.name_and_extension}')
			raise e

		paragraph_start = get_now()
		try:
			paragraphs = text.get_paragraphs(paragraph_max=paragraph_max, text_as_string=False)
		except Exception as e:
			print(f'exception raised when tried to get_paragraphs from text of {self.name_and_extension}')
			raise e

		records = []
		section_title = None
		title_font_size = None
		section_num = 0
		paragraph_num = 0
		for paragraph in paragraphs:
			if isinstance(paragraph, Title):
				section_num += 1
				section_title = paragraph.string
				title_font_size = paragraph.font_size
			elif isinstance(paragraph, Text):
				paragraph_num += 1
				for i, paragraph_text in enumerate(paragraph.paragraphs):
					record = {
						'pdf': self.name,
						'section': section_num,
						'paragraph': paragraph_num,
						'part': i + 1,
						'title': section_title,
						'title_font_size': title_font_size,
						'paragraph_font_size': paragraph.font_size,
						'text': paragraph_text,
					}
					records.append(record)

		self._elapsed['paragraph_generation'] = get_elapsed(start=paragraph_start, unit='s')
		return records

	def get_data(
			self, simplify=True, ignore_depth=True, remove_short_strings=False,
			paragraph_max=50, min_english_ratio=0, measure_time=False
	):
		"""
		:type simplify: bool
		:type ignore_depth: bool
		:type remove_short_strings: bool
		:type paragraph_max: int
		:rtype: DataFrame
		"""
		start_time = get_now()
		records = self.get_text_records(
			simplify=simplify,
			ignore_depth=ignore_depth,
			remove_short_strings=remove_short_strings,
			paragraph_max=paragraph_max,
			min_english_ratio=min_english_ratio
		)

		data = DataFrame.from_records(records)
		self._elapsed['total'] = get_elapsed(start=start_time, unit='s')

		if measure_time:
			for key, value in self._elapsed.items():
				data[f'{key}_time'] = value
		return data



from bs4 import Tag
from .Text import Text, Title
from .tag import get_text
from .tag import get_children
from .tag import get_font_size
from .tag import get_visible_border_size
from .tag import is_tag_title
from .tag import previous_tag_might_be_title
from .string import is_short_text
from .string import is_capitalized
from .string import ends_with_punctuation
from .string import get_alphabetic_ratio
from .string import get_ratio_of_english_words
from .string import get_num_words


class HtmlText:
	def __init__(
			self, tag, depth=0, strong_sep='\n', sep=' ', parent=None, is_title=False,
			font_size=None, border_size=None,
			vocabulary=None, language='english',
			index=None, log=False
	):
		"""
		:type tag: Tag
		:type depth: int
		"""
		if tag is None:
			raise

		self._do_log = log
		self._id = None

		if self._do_log:
			if parent:
				if index is None:
					raise ValueError('need an index!')
				elif parent.id is not None:
					self._id = f'{parent.id}.{index}'
				else:
					self._id = f'{index}'


				self._logs = parent._logs
				self.log('added')
				parent._children_dictionary[index] = self
			else:
				self._logs = {'first one': ['created']}

			self._children_dictionary = {}

		self._depth = max(0, depth)
		self._strong_sep = strong_sep
		self._sep = sep
		self._vocabulary = vocabulary
		self._language = language

		tag_children = get_children(tag)
		self._clean = False
		self._parent = parent
		self._tag = tag

		self._font_size = font_size or get_font_size(tag)
		self._border_size = border_size or get_visible_border_size(tag)
		self._is_table = border_size is None and self._border_size is not None
		self._is_short = False
		self._previous_might_be_title = False

		if tag_children is None:
			self._children = None
			self._string = get_text(tag, remove_non_alphabetic=is_title)
			self._is_title = is_title

			if not self._string:
				self._string = None
			else:
				self._is_short = is_short_text(self._string)

		else:
			self._string = None
			if is_tag_title(tag=tag, children=tag_children):
				self._is_title = True
			else:
				self._is_title = is_title

			self._previous_might_be_title = previous_tag_might_be_title(tag=tag, children=tag_children)

			self._children = []
			index = 0
			for child_tag in tag_children:
				self._children.append(
					HtmlText(
						tag=child_tag, depth=self._depth + 1, strong_sep=self._strong_sep, sep=self._sep,
						parent=self, is_title=self._is_title,
						font_size=self._font_size, border_size=self._border_size,
						index=index, log=log
					)
				)
				index += 1

			if not self._children:
				self._children = None
		self.clean_up(remove_tables=False)

		# if most children are short text it's a table
		if self.children:
			num_string_children = 0
			num_short_strings = 0
			for child in self.children:
				if child._string:
					num_string_children += 1
					if child._is_short:
						num_short_strings += 1

			if num_string_children > 3:
				if num_short_strings + 1 == num_string_children:
					self._is_table = True
					self.log(f'is table because num_short_strings={num_short_strings} and num_string_children={num_string_children}')
				elif num_string_children > 0 and num_short_strings / num_string_children >= 0.9:
					self._is_table = True
					self.log(f'is table because num_short_strings={num_short_strings} and num_string_children={num_string_children}')

	@property
	def id(self):
		return self._id

	def is_empty(self):
		if not self._children:
			self._children = None
		if not self._string and not self._children:
			return True
		else:
			return False

	@property
	def parent(self):
		"""
		:rtype: NoneType or HtmlText
		"""
		return self._parent

	# CLEAN UP

	def log(self, message):
		"""
		:type message: str
		"""
		if self._do_log:
			if self.id in self._logs:
				self._logs[self.id].append(message)
			else:
				self._logs[self.id] = [message]

	@property
	def logs(self):
		"""
		:rtype: dict[str, list[str]]
		"""
		return self._logs

	def log_for_all_children(self, message):
		if self._do_log:
			self.log(message=message)
			if self.children:
				for child in self.children:
					child.log_for_all_children(message=message)

	def clean_up(self, remove_tables=True):
		# removes empty children and grand children
		if self._children:
			children = self.children.copy()
			self._children = []
			for child in children:
				if remove_tables and child._is_table:
					child.log_for_all_children(message=f'table removed by {self.id}')
					continue
				if not child._clean:
					child.clean_up(remove_tables=remove_tables)
				if not child.is_empty():
					self._children.append(child)
				else:
					child.log(f'empty removed by {self.id}')

		if not self._children:
			self._children = None
			if not self._string and self.parent:
				self.parent._clean = False

		# some titles can be determined by looking at the second tag among children (which makes the first tag, a title)
		if self._children and len(self._children) > 1:
			first = self._children[0]
			second = self._children[1]
			if second._previous_might_be_title and first._string and is_capitalized(first._string):
				first._is_title = True
				first._string = get_text(first._tag, remove_non_alphabetic=True)
				second._previous_might_be_title = False

		self._clean = True

	def flatten(self, ignore_tables=True):
		self.log('flatten')
		if self._is_table and ignore_tables:
			return self
		else:
			if self.children:
				new_children = []
				for child in self.children:
					child.flatten(ignore_tables=ignore_tables)
					if child.children:
						for grand_child in child.children:
							if not grand_child._is_short:
								new_children.append(grand_child)
								grand_child._parent = self
							else:
								grand_child.log(f'short grand child "{grand_child._string}" removed by {self.id}')

						child._children = None
					else:
						if not child._is_short:
							new_children.append(child)
						else:
							child.log(f'short child "{child._string}" removed by {self.id}')
				self._children = new_children
			return self

	def simplify(
			self, ignore_depth=False,
			remove_short_strings=False,
			min_alphabetic_ratio=0.5,
			min_english_ratio=0.5,
			remove_tables=True
	):
		# merges children that are similar
		self.log('simplify')
		self.clean_up(remove_tables=remove_tables)
		if self.children is not None:
			if len(self.children) == 1:
				self.become_first_child(ignore_depth=ignore_depth)

		if self.children is not None:
			new_children = []
			for child in self.children:
				child.simplify(
					ignore_depth=ignore_depth,
					remove_short_strings=remove_short_strings,
					min_alphabetic_ratio=min_alphabetic_ratio,
					min_english_ratio=min_english_ratio
				)

			for child in self.children:
				if child.is_empty():
					child.log(f'empty removed in simplify by {self.id}')
					continue
				else:
					if len(new_children) == 0:
						new_children.append(child)
					else:
						last_child = new_children[-1]
						if last_child._is_title or child._is_title:
							new_children.append(child)
						elif last_child._string and ends_with_punctuation(last_child._string):
							new_children.append(child)
						elif last_child.is_similar_to(child, ignore_depth=ignore_depth):
							last_child.absorb(child)
						else:
							new_children.append(child)

			if remove_short_strings:
				new_children_2 = []
				for child in new_children:
					if child._is_short:
						child.log(f'short child removed in simplify by {self.id}')
					else:
						new_children_2.append(child)
			else:
				new_children_2 = new_children

			if min_alphabetic_ratio > 0:
				new_children_3 = []

				for child in new_children_2:
					if child._string is None:
						new_children_3.append(child)
					else:
						alphabetic_ratio = get_alphabetic_ratio(child._string)
						if not child._string or alphabetic_ratio > min_alphabetic_ratio:
							new_children_3.append(child)
						else:
							child.log(
								message=f'removed by {self.id} because alphabetic ratio {alphabetic_ratio} < {min_alphabetic_ratio}'
							)
			else:
				new_children_3 = new_children_2

			if min_english_ratio > 0 and self._vocabulary:
				new_children_4 = []
				for child in new_children_3:
					if child._string is None:
						new_children_4.append(child)
					else:
						ratio_of_english_words = get_ratio_of_english_words(
							child._string, vocabulary=self._vocabulary, case_sensitive=False
						)
						if not child._string or ratio_of_english_words > min_english_ratio or get_num_words(child._string) < 5:
							new_children_4.append(child)
						else:
							child.log(
								message=f'removed by {self.id} because ratio of english {ratio_of_english_words} < {min_english_ratio}'
							)
			else:
				new_children_4 = new_children_3

			self._children = new_children_4
		self.clean_up(remove_tables=remove_tables)
		return self

	# MERGING

	def is_similar_to(self, other, ignore_depth=False):
		"""
		:type other: HtmlText
		:type ignore_depth: bool
		:rtype: bool
		"""
		if self._depth != other._depth and not ignore_depth:
			return False
		if self._font_size is not None and other._font_size is not None and self._font_size < other._font_size:
			return False
		else:
			return (self._children is None and other._children is None) or (self._children is not None and other._children is not None)

	def absorb(self, other):
		"""
		:type other: HtmlText
		"""
		self.log(f'absorbs {other.id}')
		if self._is_title or other._is_title:
			raise RuntimeError(f'cannot absorb a title {other} into another {self}')

		if self._is_table and not other._is_table:
			raise RuntimeError(f'table cannot absorb non table')

		if not self._is_table and other._is_table:
			raise RuntimeError(f'table cannot be absorbed by non table')

		if self._children is None:
			self._string = self._string.strip()
			other._string = other._string.strip()
			if self._string:
				self._string += self._sep + other._string
			else:
				self._string += other._string
			other._string = None
			other.log(f'absorbed by {self.id}')
		else:
			self._children += other._children
			other._children = None
			other.log(f'children absorbed by {self.id}')

		if self._font_size and other._font_size:
			self._font_size = max(self._font_size, other._font_size)
		elif other._font_size:
			self._font_size = other._font_size

		self._clean = self._clean and other._clean
		self.parent._clean = False
		other.parent._clean = False

	def become_first_child(self, ignore_depth):
		single_child = self.children[0]
		self.log(f'becomes first child {single_child.id}')
		single_child.simplify(ignore_depth=ignore_depth)
		if single_child.children is None:
			self._string = single_child._string
			self._children = None
			single_child.log(f'taken over by parent {self.id}')
		else:
			self._children = single_child.children
			single_child.log(f'children taken over by parent {self.id}')

		self._is_title = self._is_title or single_child._is_title
		self._is_table = self._is_table or single_child._is_table
		self._border_size = self._border_size or single_child._border_size
		self._font_size = self._font_size or single_child._font_size
		self._clean = self._clean and single_child._clean

	#  CHILDREN

	def has_children(self):
		return self._children is not None

	@property
	def children(self):
		"""
		:rtype: list[HtmlText]
		"""
		return self._children

	def __getitem__(self, item):
		return self._children_dictionary[item]

	# STRING

	def get_string(self):
		if self._string is None and self._children is None:
			raise RuntimeError(f'cannot display an empty text {self.id}')
		elif self._children is None:
			if self._is_title:
				string = f'* {self._string} *'
			else:
				string = f'"{self._string}"'

			if self._font_size:
				string = f'({self._font_size}):{string}'

			if self._is_table:
				string = f'|| {string} ||'
			elif self._border_size:
				string = f'| {string} |'

			return f'{self.id} <{self._depth}> {string}'
		else:
			if any([child.has_children() for child in self.children]):
				return self._strong_sep.join([child.get_string() for child in self.children])
			else:

				return self._sep.join([child.get_string() for child in self.children])

	def get_paragraphs(self, paragraph_max=50, text_as_string=True):
		"""
		:rtype: str or list[str or list[str] or list[list]]
		"""

		if self.children:
			result = []
			for child in self.children:
				text = child.get_paragraphs(paragraph_max=paragraph_max, text_as_string=text_as_string)
				total_paragraph_font_size = 0
				num_strings = 0

				if isinstance(text, Text):
					if text_as_string:
						result += text.paragraphs
					else:
						result.append(text)
					if text.font_size:
						total_paragraph_font_size += text.font_size
						num_strings += len(text.paragraphs)
				else:
					# append the title
					result.append(text)

				# remove titles with font size smaller than average paragraph font size
				'''
				if total_paragraph_font_size > 0 and num_strings > 0:
					mean_paragraph_font_size = total_paragraph_font_size / num_strings
					result = [x for x in result if not isinstance(x, Title) or x.font_size > mean_paragraph_font_size]
				'''
			return result
		else:
			if self._string is None:
				print(f'error being raised for child of {self.parent._tag}')
				raise ValueError(f' has no string!')
			string = ' '.join(self._string.split())
			if self._is_title:
				return Title(string, font_size=self._font_size)
			else:
				return Text(
					string, font_size=self._font_size, paragraph_max=paragraph_max, language=self._language
				)

	def __str__(self):
		return self.get_string()

	def __repr__(self):
		return str(self)





from pdfminer.layout import LAParams
from pdfminer.converter import PDFPageAggregator
from pdfminer.pdfinterp import PDFResourceManager
from pdfminer.pdfinterp import PDFPageInterpreter
from pdfminer.pdfpage import PDFPage
from pdfminer.layout import LTTextBoxHorizontal


def convert_pdf_to_text(document):
    # convert all horizontal text into a lines list (one entry per line)
    # document is a file stream
    lines = []
    rsrcmgr = PDFResourceManager()
    laparams = LAParams()
    device = PDFPageAggregator(rsrcmgr, laparams=laparams)
    interpreter = PDFPageInterpreter(rsrcmgr, device)
    for page in PDFPage.get_pages(document):
            interpreter.process_page(page)
            layout = device.get_result()
            for element in layout:
                if isinstance(element, LTTextBoxHorizontal):
                    lines.extend(element.get_text().splitlines())
    return lines